#include "TT6PedestalCalculator.h"
#include <cmath>
#include <numeric>
#include <algorithm>
//
//  -- Constructor. 
//
TT6PedestalCalculator::TT6PedestalCalculator(int evnt_ini, 
                        int evnt_iter, float sig_cut) :   
                        eventsForInitialization(evnt_ini), 
                        eventsForIteration(evnt_iter),
	                cutToAvoidSignal(sig_cut),
                        numberOfEvents(0),
                        alreadyUsedEvent(false)
{
  if (0) cout << "Constructing TT6PedestalCalculator " << endl;

  thePedestal.clear();
  thePedSum.clear();
  thePedSqSum.clear();
  theEventPerStrip.clear();
  theStatus.setCalibrating();
}
//
//  -- Destructor  
//
TT6PedestalCalculator::~TT6PedestalCalculator() {
  if (0) cout << "Destructing TT6PedestalCalculator " << endl;
}
//
// -- Set Pedestal Update Status
//
void TT6PedestalCalculator::updateStatus(){
  if (theStatus.isCalibrating() && 
      numberOfEvents >= eventsForInitialization) {
    if (accumulate(thePedestal.begin(), thePedestal.end(),0.) != 0.) {
      theStatus.setUpdating();
    }
  }
}

//
// -- Initialize Pedestal Values using a set of events (eventsForInitialization)
//
void TT6PedestalCalculator::initializePedestal(ApvAnalysis::RawSignalType in) {
  unsigned int i;
  if (alreadyUsedEvent == false) {
    alreadyUsedEvent = true;
    if (numberOfEvents == 1) {
      for (i = 0; i < in.size(); i++) {
	ApvAnalysis::RawSignalType temp;
	temp.reserve(eventsForInitialization);
	theRawSignalEventStrip.insert( pair<unsigned short, ApvAnalysis::RawSignalType>(i, temp) );
      }
    }
    
    if (numberOfEvents <= eventsForInitialization) {
      for (i = 0; i < in.size(); i++) {
	theRawSignalEventStrip[i].push_back(in[i]);
      }
    }
    if (numberOfEvents == eventsForInitialization) {
      thePedestal.clear();
      theRawNoise.clear();
      for (i = 0; i < in.size(); i++) {
        ApvAnalysis::SubtractedSignalType temp;
        temp.resize(eventsForInitialization);
        copy ( theRawSignalEventStrip[i].begin(), 
               theRawSignalEventStrip[i].end(), temp.begin());
	sort(temp.begin(), temp.end());

        int len = temp.size();
        int nTrunc = int(len*0.20);
	int rlen = len-nTrunc;

        double sumVal = 0.0;
        double sqSumVal = 0.0;
 
        sumVal   = accumulate(temp.begin(), (temp.end()-nTrunc), sumVal);
        sqSumVal = inner_product(temp.begin(), (temp.end()-nTrunc), 
                                    temp.begin(), sqSumVal);

	double avVal    = (rlen) ? sumVal/rlen   : 0.0;
	double sqAvVal  = (rlen) ? sqSumVal/rlen : 0.0;
	double rmsVal   = (sqAvVal - avVal*avVal > 0.0) 
             ? sqrt(sqAvVal - avVal*avVal) : 0.0;	
	if (0) cout << "TT6PedestalCalculator ::initializePedestal " 
                    << i << " " << sumVal << " " 
                    << sqSumVal << " "
                    << rlen << " " << avVal << " " << sqAvVal << " "  
                    << rmsVal << endl; 
        thePedestal.push_back(static_cast<float>(avVal));
        theRawNoise.push_back(static_cast<float>(rmsVal));
      }
      theRawSignalEventStrip.clear();      
    }
  }
}
//
// -- Calculate and Update (when needed) Pedestal Values
//
void TT6PedestalCalculator::updatePedestal(ApvAnalysis::RawSignalType in) {
  if (!theStatus.isUpdating()) return;
  if (alreadyUsedEvent == false) {
    alreadyUsedEvent = true;
    if (((numberOfEvents-eventsForInitialization)%eventsForIteration) == 1) {

      thePedSum.clear();
      thePedSqSum.clear();
      theEventPerStrip.clear();

      thePedSum.resize(in.size(), 0.0);
      thePedSqSum.resize(in.size(), 0.0);
      theEventPerStrip.resize(in.size(), 0);
    }
    unsigned int i;
    for (i = 0; i < in.size(); i++) {
      if (fabs(in[i]-thePedestal[i]) < cutToAvoidSignal*theRawNoise[i]) {
        thePedSum[i]   += in[i];
        thePedSqSum[i] += in[i]*in[i];
        theEventPerStrip[i]++;
      }
    }
    if (((numberOfEvents-eventsForInitialization) % eventsForIteration) == 0) {

      thePedestal.clear();
      theRawNoise.clear();

      for (i = 0; i < in.size(); i++) {
        double avVal   = (theEventPerStrip[i]) 
          ? thePedSum[i]/theEventPerStrip[i]:0.0;
        double sqAvVal = (theEventPerStrip[i]) 
          ? thePedSqSum[i]/theEventPerStrip[i]:0.0;
        double rmsVal  =  sqrt(sqAvVal - avVal*avVal);

	if (0) cout << "TT6PedestalCalculator::updatePedestal " 
                    << thePedSum[i]   << " " 
                    << thePedSqSum[i] << " "
                    << avVal << " " << sqAvVal << " "  
                    << rmsVal << endl; 
        thePedestal.push_back(static_cast<float>(avVal));
        theRawNoise.push_back(static_cast<float>(rmsVal));
      }
    }
  }
}
//
// -- doIt
//
ApvAnalysis::SubtractedSignalType 
TT6PedestalCalculator::doIt(ApvAnalysis::RawSignalType in) {
  
  ApvAnalysis::SubtractedSignalType temp;
  temp.resize(in.size(), 0.0);

  if (alreadyUsedEvent == false) {
    if (theStatus.isCalibrating()) {
       initializePedestal(in);
    } else if (theStatus.isUpdating()) {
      updatePedestal(in);
    }  
    if (theStatus.isCalibrating() && numberOfEvents >= eventsForInitialization) updateStatus();
  }
  if (theStatus.isUpdating()) {
    transform(in.begin(), in.end(), thePedestal.begin(), 
                           temp.begin(), minus<float>());   
  }
  return temp;
}
//
// Define New Event
// 
void TT6PedestalCalculator::newEvent(){
  alreadyUsedEvent = false;
  numberOfEvents++;
}
